// Copyright (c) 2014, The Shelf Bind project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf.bind.test;

import 'package:shelf_bind/shelf_bind.dart' as sb;
import 'package:shelf/shelf.dart' as s;
import 'package:shelf_route/shelf_route.dart' as sr;


import 'package:unittest/unittest.dart';
import 'package:mock/mock.dart';
import 'dart:convert';
import 'dart:async';

typedef _PersonHandler1(AbstractPerson person);
typedef _PersonHandler2(AbstractPerson person, s.Request request);

abstract class PersonHandler1Class {
  call(AbstractPerson person) { return null; }
}

abstract class PersonHandler2Class {
  call(AbstractPerson person, s.Request request) { return null; }
}

class MockPersonHandler1 extends Mock implements PersonHandler1Class {
  noSuchMethod(Invocation invocation) => super.noSuchMethod(invocation);
}

class MockPersonHandler2 extends Mock implements PersonHandler2Class {
  noSuchMethod(Invocation invocation) => super.noSuchMethod(invocation);
}


main() {
  MockPersonHandler1 mockHandler1;
  MockPersonHandler2 mockHandler2;

  _PersonHandler1 singleArgMockHandler() =>
      (AbstractPerson person) => mockHandler1.call(person);

  _PersonHandler2 twoArgMockHandler() =>
      (AbstractPerson person, s.Request request) =>
        mockHandler2.call(person, request);


  setUp(() {
    mockHandler1 = new MockPersonHandler1();
    mockHandler2 = new MockPersonHandler2();
  });

  group('bind', () {
    void runTests(Type personType, sb.BindMode bindMode) {
      s.Handler _bindHandlerFor(wrappedHandler) =>
          sb.bind(personType, {'name': #name},
          wrappedHandler, bindMode: bindMode);

      s.Handler _singleArgBindHandler() =>
          _bindHandlerFor(singleArgMockHandler());
      s.Handler _twoArgBindHandler() => _bindHandlerFor(twoArgMockHandler());


      test('should bind to variable in uri path', () {
        mockHandler2.when(callsTo('call')).alwaysReturn(new s.Response.ok(''));
        final Uri uri = Uri.parse('http://localhost/person/fred');
        final bindHandler = _twoArgBindHandler();
        final routeHandler = (sr.router()..get('/person/{name}', bindHandler))
            .handler;

        final request = new s.Request('GET', uri);
        routeHandler(request);

        mockHandler2.calls('call', personWithName(equals('fred')))
            .verify(happenedOnce);
      });

      test('should bind to variable in query parameter', () {
        mockHandler2.when(callsTo('call')).alwaysReturn(new s.Response.ok(''));
        final Uri uri = Uri.parse('http://localhost/person?name=fred');
        final bindHandler = _twoArgBindHandler();
        final routeHandler = (sr.router()..get('/person{?name}', bindHandler))
            .handler;

        final request = new s.Request('GET', uri);
        routeHandler(request);

        mockHandler2.calls('call', personWithName(equals('fred')))
            .verify(happenedOnce);
      });

      test('should support functions with only one argument', () {
        mockHandler1.when(callsTo('call')).alwaysReturn(new s.Response.ok(''));

        final Uri uri = Uri.parse('http://localhost/person/fred');
        final bindHandler = _singleArgBindHandler();
        final routeHandler = (sr.router()..get('/person/{name}', bindHandler))
            .handler;

        final request = new s.Request('GET', uri);
        routeHandler(request);

        mockHandler1.calls('call', personWithName(equals('fred')))
            .verify(happenedOnce);
      });

      test('should convert non Response values to json', () {
        mockHandler1.when(callsTo('call'))
            .alwaysReturn(new ConstructorPerson('barney'));
        final Uri uri = Uri.parse('http://localhost/person/fred');
        final bindHandler = _singleArgBindHandler();
        final routeHandler = (sr.router()..get('/person/{name}', bindHandler))
            .handler;

        final request = new s.Request('GET', uri);
        final responseFuture = routeHandler(request);
        responseFuture.then((r) {
          expect(r, new isInstanceOf<s.Response>());
          s.Response response = r;
          expect(response.headers['content-type'], equals('application/json'));
          response.readAsString().then((str) {
            expect(JSON.decode(str)['name'], equals('barney'));
          });
        });
        expect(responseFuture, completes);
      });

    }

    group('in constructor mode', () {
      runTests(ConstructorPerson, sb.BindMode.CONSTRUCTOR);
    });

    group('in property mode with variable', () {
      runTests(FieldPerson, sb.BindMode.PROPERTY);
    });

    group('in property mode with setter', () {
      runTests(SetterPerson, sb.BindMode.PROPERTY);
    });

  });

  group('bindJsonBody', () {
    s.Request _createRequest(Uri uri) {
      final json = JSON.encode(new ConstructorPerson('fred').toJson());
      final body = new Future.sync(() => json.codeUnits).asStream();
      return new s.Request('POST', uri, body: body);
    }

    test('should bind to body', () {
      mockHandler2.when(callsTo('call')).alwaysReturn(new s.Response.ok(''));
      final Uri uri = Uri.parse('http://localhost/person');
      final bindHandler = sb.bindJsonBody(ConstructorPerson,
          twoArgMockHandler());
      final routeHandler = (sr.router()..post('/person', bindHandler))
          .handler;

      var request = _createRequest(uri);
      final responseFuture = new Future.sync(() => routeHandler(request))
      .then((_) {
        mockHandler2.calls('call', personWithName(equals('fred')))
            .verify(happenedOnce);
      });
      expect(responseFuture, completes);
    });

    test('should support functions with only one argument', () {
      mockHandler1.when(callsTo('call')).alwaysReturn(new s.Response.ok(''));

      final Uri uri = Uri.parse('http://localhost/person');
      final bindHandler = sb.bindJsonBody(ConstructorPerson,
          singleArgMockHandler());
      final routeHandler = (sr.router()..post('/person', bindHandler))
          .handler;

      final request = _createRequest(uri);
      final responseFuture = new Future.sync(() => routeHandler(request))
      .then((_) {
        mockHandler1.calls('call', personWithName(equals('fred')))
            .verify(happenedOnce);
      });
      expect(responseFuture, completes);
    });

    test('should convert non Response values to json', () {
      mockHandler2.when(callsTo('call'))
          .alwaysReturn(new ConstructorPerson('barney'));

      final Uri uri = Uri.parse('http://localhost/person');
      final bindHandler = sb.bindJsonBody(ConstructorPerson,
          twoArgMockHandler());
      final routeHandler = (sr.router()..post('/person', bindHandler))
          .handler;

      final request = _createRequest(uri);

      final responseFuture = routeHandler(request);
      responseFuture.then((r) {
        expect(r, new isInstanceOf<s.Response>());
        s.Response response = r;
        expect(response.headers['content-type'], equals('application/json'));
        response.readAsString().then((str) {
          expect(JSON.decode(str)['name'], equals('barney'));
        });
      });
      expect(responseFuture, completes);
    });

  });
}


s.Response _handlePerson(AbstractPerson person, s.Request request) {
  print(person);
  return new s.Response.ok('yippee - ${person.toJson()}');
}

Matcher personWithName(matcher) => _personMatcher("name", matcher,
    (AbstractPerson p) => p.name);

Matcher _personMatcher(String fieldName, matcher, Getter getter)
    => fieldMatcher("Person", fieldName, matcher, getter);

Matcher fieldMatcher(String className, String fieldName, matcher,
                     Getter getter) =>
    new FieldMatcher(className, fieldName, matcher, getter);


typedef Getter(object);

class FieldMatcher extends CustomMatcher {
  final Getter getter;

  FieldMatcher(String className, String fieldName, matcher, this.getter)
      : super("$className with $fieldName that", fieldName, matcher);

  featureValueOf(actual) => getter(actual);
}


abstract class AbstractPerson {
  String get name;

  Map toJson() => { 'name': name };

  String toString() => '$runtimeType[name: $name]';
}

class ConstructorPerson extends AbstractPerson {
  final String name;

  ConstructorPerson(this.name);

  ConstructorPerson.build({String name}) : this(name);

  ConstructorPerson.fromJson(Map json) : this(json['name']);
}

class FieldPerson extends AbstractPerson {
  String name;

  FieldPerson();

  FieldPerson.fromJson(Map json) : this.name = json['name'];
}

class SetterPerson extends AbstractPerson {
  String _name;

  void set name(String name) {
    _name = name;
  }

  String get name => _name;

  SetterPerson();

  SetterPerson.fromJson(Map json) { this.name = json['name']; }
}